use std::sync::mpsc::{Receiver, Sender, TryRecvError};
use std::time::Duration;

use ::spawner::Spawner;
use ::task::Task;
use ::task_state::TaskState;
use ::task_state_machine::TaskStateMachine;
use ::thread::Thread;
use ::thread_state::ThreadState;



///
pub struct BlockingThread
{
   ///
   state: ThreadState,

   ///
   current_task: Box<TaskStateMachine>,

   ///
   shutdown_receiver: Receiver<bool>,

   ///
   state_sender: Sender<ThreadState>,

   ///
   continue_running: bool,

   ///
   spawner: Spawner
}



impl BlockingThread
{
   /// 
   pub fn new(task: Box<TaskStateMachine>,
              shutdown_r: Receiver<bool>,
              state_s: Sender<ThreadState>)
      -> BlockingThread
   {
      BlockingThread
      {
         state: ThreadState::Starting,
         current_task: task,
         shutdown_receiver: shutdown_r,
         state_sender: state_s,
         continue_running: true,
         spawner: Spawner::new()
      }
   }

   ///
   fn change_state(&mut self, new_state: ThreadState)
   {
      self.state = new_state;
      match self.state_sender.send(self.state)
      {
         Ok(_) =>
         {
         }

         Err(error) =>
         {
            // We lost our connection to the
            // state channel.
            warn!("{}", error);
         }
      }
   }

   /// Queues a task to be processed. For a
   /// blocking thread this needs to queue a
   /// new blocking thread task.
   fn queue_task(&mut self, task: Box<Task>)
   {
   }

   /// Check for any new thread shutdown messages.
   fn process_shutdown_messages(&mut self)
   {
      let mut check_messages: bool;

      // Loop through all the messages in the
      // receivers queue.
      check_messages = true;
      while check_messages == true
      {
         match self.shutdown_receiver.try_recv()
         {
            Ok(val) =>
            {
               // Found a message.
               self.continue_running = !val;
            }

            Err(error) =>
            {
               // Determine the kind of error we received.
               match error
               {
                  TryRecvError::Empty =>
                  {
                     // No messages to handle.
                     check_messages = false;
                     debug!("There were no shutdown messages.");
                  }

                  TryRecvError::Disconnected =>
                  {
                     // We lost our connection to the
                     // shutdown channel.
                     // TODO: Handle this poisoning correctly.
                     warn!("{}", error);
                  }
               }
            }
         }
      }
   }
}

impl Thread for BlockingThread
{
   ///
   fn process(&mut self)
   {
      // Run this thread until the scheduler decides
      // to shut it down.
      self.change_state(ThreadState::Idle);
      self.change_state(ThreadState::Processing);
      while self.continue_running == true
      {
         // Process the task this thread is
         // currently working on.
         self.current_task.process(&mut self.spawner);

         match self.current_task.get_state()
         {
            TaskState::Finished =>
            {
               self.continue_running = false;
            }

            _ =>
            {
            }
         }

         // Sleep the thread so that other threads
         // get a chance to run.
         ::std::thread::sleep(Duration::new(0, 100));

         // Check to see if this thread should be shutdown.
         if self.continue_running == true
         {
            self.process_shutdown_messages();
         }
      }

      // This thread is finished.
      println!("Shutting down thread.");
      self.change_state(ThreadState::Finished);
   }
}
