use std::collections::binary_heap::BinaryHeap;
use std::ops::DerefMut;
use std::sync::{Arc, Mutex};
use std::sync::mpsc::{Receiver, Sender, TryRecvError};
use std::time::Duration;

use ::spawner::Spawner;
use ::task::Task;
use ::task_state::TaskState;
use ::task_state_machine::TaskStateMachine;
use ::thread::Thread;
use ::thread_state::ThreadState;



///
pub struct ComputeThread
{
   ///
   state: ThreadState,

   ///
   current_task: Option<Box<TaskStateMachine>>,

   ///
   task_queue: Arc<Mutex<BinaryHeap<Box<TaskStateMachine>>>>,

   ///
   shutdown_receiver: Receiver<bool>,

   ///
   state_sender: Sender<ThreadState>,

   ///
   continue_running: bool,

   ///
   spawner: Spawner
}



impl ComputeThread
{
   /// 
   pub fn new(queue: Arc<Mutex<BinaryHeap<Box<TaskStateMachine>>>>,
              shutdown_r: Receiver<bool>,
              state_s: Sender<ThreadState>)
      -> ComputeThread
   {
      ComputeThread
      {
         state: ThreadState::Starting,
         current_task: None,
         task_queue: queue,
         shutdown_receiver: shutdown_r,
         state_sender: state_s,
         continue_running: true,
         spawner: Spawner::new()
      }
   }

   ///
   fn change_state(&mut self, new_state: ThreadState)
   {
      self.state = new_state;
      match self.state_sender.send(self.state)
      {
         Ok(_) =>
         {
         }

         Err(error) =>
         {
            // We lost our connection to the
            // state channel.
            warn!("{}", error);
         }
      }
   }

   ///
   fn retrieve_task(&mut self)
   {
      let task: Option<Box<TaskStateMachine>>;

      // There is nothing to do if this thread already
      // has a task to work on.
      if self.current_task.is_some() == true
      {
         return;
      }

      // This thread does not have a current task.
      // Get another task to work on from the Queue.
      match self.task_queue.lock()
      {
         Ok(ref mut guard) =>
         {
            self.current_task = guard.deref_mut().pop();

            match self.current_task
            {
               Some(_) =>
               {
                  debug!("Received a task.");
               }

               None =>
               {
                  debug!("No new task to process.");
               }
            }
         }

         Err(error) =>
         {
            error!("{}", error);
         }
      }
   }

   /// Queues a task to be processed.
   fn queue_task(&mut self, task: Box<TaskStateMachine>)
   {
      // Just add the task to the queue.
      match self.task_queue.lock()
      {
         Ok(ref mut guard) =>
         {
            guard.deref_mut().push(task);
         }

         Err(error) =>
         {
            error!("{}", error);
         }
      }
   }

   /// Check for any new thread shutdown messages.
   fn process_shutdown_messages(&mut self)
   {
      let mut check_messages: bool;

      // Loop through all the messages in the
      // receivers queue.
      check_messages = true;
      while check_messages == true
      {
         match self.shutdown_receiver.try_recv()
         {
            Ok(val) =>
            {
               // Found a message.
               self.continue_running = !val;
            }

            Err(error) =>
            {
               // Determine the kind of error we received.
               match error
               {
                  TryRecvError::Empty =>
                  {
                     // No messages to handle.
                     check_messages = false;
                     debug!("There were no shutdown messages.");
                  }

                  TryRecvError::Disconnected =>
                  {
                     // We lost our connection to the
                     // shutdown channel.
                     // TODO: Handle this poisoning correctly.
                     warn!("{}", error);
                  }
               }
            }
         }
      }
   }
}

impl Thread for ComputeThread
{
   ///
   fn process(&mut self)
   {
      let mut check_messages: bool;

      // Run this thread until the scheduler decides
      // to shut it down.
      while self.continue_running == true
      {
         match self.state
         {
            ThreadState::Starting =>
            {
               // The starting state can only lead to the
               // idle state, and does so directly.
               self.change_state(ThreadState::Idle);
            }

            ThreadState::Idle =>
            {
               let mut is_valid_task: bool;

               // Check to see if we have a task.
               is_valid_task = self.current_task.is_some();
               if is_valid_task == true
               {
                  // Why are we idling? We should not get here.
                  debug!("Thread is idling when it should be processing.");
                  self.change_state(ThreadState::Processing);
               }
               else
               {
                  // We need to check if there is a task that
                  // we should be processing.
                  self.retrieve_task();
                  is_valid_task = self.current_task.is_some();
                  if is_valid_task == true
                  {
                     // We got a new task to process,
                     // so switch to the Processing state.
                     self.change_state(ThreadState::Processing);
                  }
               }

               // Check to see if this thread should be shutdown.
               self.process_shutdown_messages();
            }

            ThreadState::Processing =>
            {
               let mut switch_to_idle: bool;
               let mut is_task_finished: bool;
               let mut has_spawned_children: bool;
               let mut tasks: Vec<Box<TaskStateMachine>>;
               let mut child_task: Option<Box<TaskStateMachine>>;

               // Check to see if we have a task. The match statement
               // makes this easier to handle, but because it borrows
               // the current task we need to mark the action to take
               // and then handle it outside of the match statement
               // if it changes the current task.
               switch_to_idle = false;
               is_task_finished = false;
               has_spawned_children = false;
               match self.current_task
               {
                  Some(ref mut task) =>
                  {
                     // Process the task this thread is
                     // currently working on.
                     is_task_finished = task.process(&mut self.spawner);

                     // If the task is not finished then
                     // check to see if any child tasks
                     // were spawned.
                     if is_task_finished == false
                     {
                        if self.spawner.has_child_tasks() == true
                        {
                           has_spawned_children = true;
                        }
                     }
                  }

                  None =>
                  {
                     // Why are we processing? We should not get here.
                     debug!("Thread is processing when it should be idling.");
                     switch_to_idle = true;
                  }
               }

               if switch_to_idle == true
               {
                  self.change_state(ThreadState::Idle);
               }

               // If there are any sibling tasks that were
               // spawned then add them to the queue.
               if self.spawner.has_sibling_tasks() == true
               {
                  tasks = self.spawner.retrieve_sibling_tasks();
                  while tasks.is_empty() == false
                  {
                     child_task = tasks.pop();
                     match child_task
                     {
                        Some(task) =>
                        {
                           self.queue_task(task);
                        }

                        None =>
                        {
                           debug!("Queueing a sibling task that \
                                   did not exist.");
                        }
                     }
                  }
               }

               // Check for any actions that need to be taken.
               if is_task_finished == true
               {
                  // If the task was finished then it can be
                  // dropped and this thread can go to an idle state.
                  self.current_task = None;
                  self.change_state(ThreadState::Idle);
               }
               else if has_spawned_children == true
               {
                  // The current task is not finished so
                  // queue any child tasks that were spawned.
                  tasks = self.spawner.retrieve_child_tasks();
                  while tasks.is_empty() == false
                  {
                     child_task = tasks.pop();
                     match child_task
                     {
                        Some(task) =>
                        {
                           self.queue_task(task);
                        }

                        None =>
                        {
                           debug!("Queueing a child task that did not exist.");
                        }
                     }
                  }
               }

               // Check to see if this thread should be shutdown.
               self.process_shutdown_messages();
            }

            ThreadState::Finished =>
            {
               let task: Option<Box<TaskStateMachine>>;

               // The thread has been put into a finished state.
               // This means that the thread should exit its loop.
               info!("Shutting down thread.");
               self.continue_running = false;

               // Also, if the thread is currently working on
               // an unfinished task then we should requeue it.
               task = self.current_task.take();
               self.current_task = None;
               match task
               {
                  Some(old_task) =>
                  {
                     if old_task.get_state() != TaskState::Finished
                     {
                        self.queue_task(old_task);
                     }
                  }

                  None =>
                  {
                  }
               }
            }
         }

         // Sleep the thread so that other threads
         // get a chance to run 
         ::std::thread::sleep(Duration::new(0, 100));
      }
   }
}
