use ::mortality::Mortality;
use ::priority::Priority;
use ::spawner::Spawner;
use ::task::Task;
use ::task_state::TaskState;
use ::task_state_machine::TaskStateMachine;



///
pub struct TaskContainer<T>
{
   ///
   state: TaskState,

   ///
   task: T
}



impl<T> TaskContainer<T> where T: Task
{
   ///
   pub fn new(decorated_task: T) -> TaskContainer<T>
   {
      TaskContainer
      {
         state: TaskState::Starting,
         task: decorated_task
      }
   }
}

impl<T> Task for TaskContainer<T> where T: Task
{
   fn get_name(&self) -> &str
   {
      // Just return the contained Task's name.
      self.task.get_name()
   }

   fn get_mortality(&self) -> Mortality
   {
      // Just return the contained Task's Mortality.
      self.task.get_mortality()
   }

   fn get_priority(&self) -> Priority
   {
      // Just return the contained Task's Priority.
      self.task.get_priority()
   }

   fn reset(&mut self)
   {
      debug!("{}: Reseting.", self.get_name());

      // Set the state back to the starting state.
      self.state = TaskState::Starting;

      // Reset the contained Task.
      self.task.reset();
   }

   fn process(&mut self, spawner: &mut Spawner) -> bool
   {
      match self.state
      {
         TaskState::Starting =>
         {
            debug!("{}: Switching from Starting to Processing.",
                   self.get_name());

            // Move directly to the Processing state.
            self.state = TaskState::Processing;

            // Return that this task is not finished.
            false
         }

         TaskState::Processing =>
         {
            debug!("{}: Processing task.", self.get_name());

            // Process the contained Task.
            if self.task.process(spawner) == true
            {
               // The task is finished processing.
               self.state = TaskState::Finished;
            }
            else
            {
               // The task is still running, so check to see
               // if any child tasks were spawned.
               if spawner.has_child_tasks() == true
               {
                  // There are child tasks, so change the state
                  // to be waiting while the children are processed.
                  self.state = TaskState::Waiting;
               }
            }

            // Return that this task is not finished.
            false
         }

         TaskState::Waiting =>
         {
            debug!("{}: Waiting for child Tasks to complete.",
                   self.get_name());

            // Return that this task is not finished.
            false
         }

         TaskState::Finished =>
         {
            debug!("{}: Task finished.", self.get_name());

            // Return that this task is finished.
            true
         }

         TaskState::Unknown =>
         {
            warn!("{}: Trying to process a task in an Unknown state.",
                  self.get_name());

            // The task is in an Unknown state, return that
            // it is not finished. The Unknown state
            // should be tested for to make sure that this
            // task is no longer processed.
            false
         }
      }
   }
}

impl<T> TaskStateMachine for TaskContainer<T> where T: Task
{
   ///
   fn get_state(&self) -> TaskState
   {
      //
      self.state
   }
}
