/// The different states a Thread can go through during its lifetime.
#[derive(Clone, Copy, PartialEq, PartialOrd, Eq, Ord)]
pub enum ThreadState
{
   /// The state that every Thread starts in.
   ///
   /// The STARTING state can only lead to the IDLE state.
   Starting,

   /// The state that the Thread is in when not processing.
   /// During this state the Thread will be looking for new tasks
   /// to work on, or if the Thread needs to be shutdown.
   ///
   /// The IDLE state can only lead to the PROCESSING or FINISHED state.
   Idle,

   /// The state a Thread is in when it is processing a Task.
   ///
   /// The PROCESSING state can only lead to the IDLE or FINISHED state.
   Processing,

   /// The state a Thread will be placed in when it is FINISHED running
   /// and is ready to be destroyed.
   ///
   /// The FINISHED state cannot lead to any states and shows that the
   /// Thread is shutdown and ready to be destroyed.
   Finished
}



impl ThreadState
{
   /// Get a str representation of this variant.
   pub fn to_str(&self) -> &'static str
   {
      match *self
      {
         ThreadState::Starting =>
         {
            "Starting"
         }

         ThreadState::Idle =>
         {
            "Idle"
         }

         ThreadState::Processing =>
         {
            "Processing"
         }

         ThreadState::Finished =>
         {
            "Finished"
         }
      }
   }

   /// Get a String representation of this variant.
   pub fn to_string(&self) -> String
   {
      String::from(self.to_str())
   }
}

impl ::std::fmt::Debug for ThreadState
{
   fn fmt(&self, f: &mut ::std::fmt::Formatter) -> ::std::fmt::Result
   {
      write!(f, "{}", self.to_str())
   }
}

impl ::std::fmt::Display for ThreadState
{
   fn fmt(&self, f: &mut ::std::fmt::Formatter) -> ::std::fmt::Result
   {
      write!(f, "{}", self.to_str())
   }
}
